// Opciones para el QUERY

// Clause Where

// Crea los querys
class QueryBuilder {
  _columns = '*';
  _whereClause = '';
  _whereArgs = [];
  _orderBy = '';
  _limit = undefined;
  _page = undefined;
  constructor(db, tableName) {
    this.db = db;
    this.tableName = tableName;
  }

  // SELECT
  static query(tableName, options) {
    let sqlParts = ['SELECT', options.columns ?? "*", 'FROM', tableName];
    if (options.where) {
      sqlParts.push("WHERE");
      sqlParts.push(options.where.clause);
    }
    if (options.order) {
      sqlParts.push("ORDER BY");
      sqlParts.push(options.order);
    }
    if (options.limit) {
      sqlParts.push("LIMIT");
      sqlParts.push(options.limit);
      if (options.page) {
        sqlParts.push("OFFSET");
        sqlParts.push(options.limit * (options.page - 1));
      }
    }
    return sqlParts.filter(p => p !== '').join(' ');
  }

  // Creates the "INSERT" sql statement
  static create(tableName, object) {
    const keys = Object.keys(object);
    const columns = keys.join(', ');
    const values = keys.map(() => '?').join(', ');
    return `INSERT INTO ${tableName} (${columns}) VALUES (${values});`;
  }

  // Creates the "INSERT" sql statement
  static createArray(tableName, array) {
    if (array.length === 0) {
      throw new Error('array is empty');
    }
    const keys = Object.keys(array[0]);
    const columns = keys.join(', ');

    //const values = objects.map(() => `(${keys.map(() => '?').join(', ')})`).join(', ');
    const values = array.map(() => {
      const placeholders = keys.map(() => '?').join(', ');
      return `(${placeholders})`;
    }).join(', ');
    return `INSERT INTO ${tableName} (${columns}) VALUES ${values};`;
  }

  // Creates the "Update" sql statement
  static update(tableName, object, whereClause) {
    // Extrae el valor de "id" y crea un nuevo objeto sin ese dato
    //const { id, ...props } = object
    const values = Object.keys(object).map(k => `${k} = ?`).join(', ');
    return `UPDATE ${tableName} SET ${values} WHERE ${whereClause};`;
  }

  // Creates the "DELETE" sql statement
  static destroy(tableName, whereClause) {
    return `DELETE FROM ${tableName} WHERE ${whereClause};`;
  }

  // Creates the "DELETE ALL" sql statement
  static destroyAll(tableName) {
    return `DELETE FROM ${tableName};`;
  }

  /**
   * Compila una clausula where
   * 
   * @param {any[]} array [1,3,4,2]
   * @returns (?, ?, ?, ?)
   */
  static whereClausePlaceholders(array) {
    var args = [];
    for (let i = 0; i < array.length; i++) {
      args.push("?");
    }
    return "(" + args.join(", ") + ")";
  }

  /**
   * @param sqlString "valor con 'comillas'"
   * @returns "'valor con \'comillas\''"
   */
  static escapeSqlString(sqlString) {
    let escape = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : '\'';
    var sb = [];
    sb.push(escape);
    if (sqlString.indexOf(escape) != -1) {
      var length = sqlString.length;
      for (var i = 0; i < length; i++) {
        var c = sqlString.charAt(i);
        if (c == escape) {
          sb.push(escape);
        }
        sb.push(c);
      }
    } else {
      sb.push(sqlString);
    }
    sb.push(escape);
    return sb.join("");
  }
  select(columns) {
    this._columns = columns;
    return this;
  }
  where(clause, args) {
    this._whereClause = clause;
    this._whereArgs = args;
    return this;
  }
  orderBy(orderBy) {
    this._orderBy = orderBy;
    return this;
  }
  limit(limit) {
    this._limit = limit;
    return this;
  }
  page(page) {
    this._page = page;
    return this;
  }

  /** 
   * SELECT
   * 
   * @returns array de registros
   */
  get() {
    const sql = QueryBuilder.query(this.tableName, {
      columns: this._columns,
      where: {
        clause: this._whereClause,
        args: this._whereArgs
      },
      order: this._orderBy,
      limit: this._limit,
      page: this._page
    });
    return this.db.executeSql(sql, this._whereArgs).then(result => result.rows);
  }

  /** 
   * INSERT
   * 
   * @param {any} values
   * @returns insertId
   */
  insert(values) {
    var sql = QueryBuilder.create(this.tableName, values);
    const params = Object.values(values);
    return this.db.executeSql(sql, params).then(result => result.insertId ?? -1);
  }

  /** 
   * INSERT
   * 
   * @param {any[]} array
   * @returns result
   */
  insertArray(array) {
    const sql = QueryBuilder.createArray(this.tableName, array);
    const params = [];
    array.forEach(obj => {
      const values = Object.values(obj);
      values.forEach(val => {
        params.push(val);
      });
    });
    return this.db.executeSql(sql, params);
  }

  /** 
   * UPDATE
   * 
   * @param {any} values
   * @returns rowsAffected
   */
  update(values) {
    const sql = QueryBuilder.update(this.tableName, values, this._whereClause);
    const params = Object.values(values);
    const whereArgs = this._whereArgs ?? [];
    return this.db.executeSql(sql, [...params, ...whereArgs]).then(result => result.rowsAffected);
  }

  /** 
   * DELETE
   * 
   * @returns rowsAffected
   */
  delete() {
    const sql = QueryBuilder.destroy(this.tableName, this._whereClause);
    return this.db.executeSql(sql, this._whereArgs).then(result => result.rowsAffected);
  }
}
export default QueryBuilder;
//# sourceMappingURL=QueryBuilder.js.map